# coding=utf-8
from __future__ import unicode_literals

import contextlib
import os
import shutil
import tempfile

import pytest
from django.core.files import File
from django.utils.crypto import get_random_string

from test_app.app.models import ModelWithFileField

try:
    from tempfile import TemporaryDirectory
except ImportError:
    from setuptools.py31compat import TemporaryDirectory


@contextlib.contextmanager
def make_temp_directory():
    temp_dir = tempfile.mkdtemp()
    yield temp_dir
    shutil.rmtree(temp_dir)


@pytest.mark.usefixtures("gcs_settings")
class TestApp:
    TEST_FILE_NAME = "test_file_" + get_random_string(6)
    TEST_FILE_NAME_UNICODE = "test_file_陰陽_" + get_random_string(6)
    TEST_FILE_CONTENT = "Brathähnchen".encode("utf8")

    @pytest.mark.django_db
    def test_manual_upload(self):
        model_instance = ModelWithFileField()

        with tempfile.TemporaryFile() as testfile:
            testfile.write(self.TEST_FILE_CONTENT)
            testfile.seek(0)

            model_instance.file.save(
                    self.TEST_FILE_NAME,
                    File(testfile)
            )

        model_instance.save()

        model_instance = ModelWithFileField.objects.first()

        with model_instance.file as f:
            assert f.read() == self.TEST_FILE_CONTENT

    @pytest.mark.django_db
    def test_http_upload(self, client):
        """
        :type client: django.test.Client
        """

        with TemporaryDirectory() as tmpdir:
            for filename in [self.TEST_FILE_NAME, self.TEST_FILE_NAME_UNICODE]:
                with open(os.path.join(tmpdir, filename), 'w+b') as testfile:
                    testfile.write(self.TEST_FILE_CONTENT)
                    testfile.seek(0)

                    r = client.post('/upload', {'file': testfile})

                    assert 302 == r.status_code

        assert ModelWithFileField.objects.count() == 2

    @pytest.mark.django_db
    def test_file_access(self, client):
        """
        :type client: django.test.Client
        """
        model_instance = ModelWithFileField()

        with tempfile.TemporaryFile() as testfile:
            testfile.write(self.TEST_FILE_CONTENT)
            testfile.seek(0)

            model_instance.file.save(
                    self.TEST_FILE_NAME,
                    File(testfile)
            )

        model_instance.save()

        model_instance = ModelWithFileField.objects.all()[0]

        r = client.get('/file/%s' % model_instance.pk)

        assert 200 == r.status_code and \
            self.TEST_FILE_CONTENT == r.content



# coding=utf-8
from __future__ import unicode_literals

import os
import string

import pytest
from django.utils.crypto import get_random_string
from django.utils import six

from django_gcloud_storage import DjangoGCloudStorage
from google.cloud.storage import Bucket

from helpers import upload_test_file

DEFAULT_CREDENTIALS_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "..",
    "test-credentials.json"
)
DEFAULT_BUCKET_LOCATION="europe-west3"
TEST_FILE_PATH = "general_test_file"
TEST_FILE_CONTENT = "Brathähnchen".encode("utf8")

def pytest_addoption(parser):
    parser.addoption("--gcs-credentials-file",
                     action="store",
                     default=DEFAULT_CREDENTIALS_PATH,
                     help="Defaults to PROJECT_DIR/test-credentials.json")
    parser.addoption("--gcs-project-name", action="store")
    parser.addoption("--gcs-bucket-location", action="store",
                     default=DEFAULT_BUCKET_LOCATION,
                     help="Defaults to " + DEFAULT_BUCKET_LOCATION)


@pytest.fixture(scope="module")
def storage(request):
    # create a random test bucket name
    bucket_name = "test_bucket_" + get_random_string(6, string.ascii_lowercase)

    storage = DjangoGCloudStorage(
        project=request.config.getoption("--gcs-project-name"),
        bucket=bucket_name,
        credentials_file_path=request.config.getoption("--gcs-credentials-file")
    )

    # Make sure the bucket exists
    bucket = Bucket(storage.client, bucket_name)
    bucket.location = request.config.getoption("--gcs-bucket-location")
    bucket.create()

    yield storage

    storage.bucket.delete_blobs(storage.bucket.list_blobs())

    storage.bucket.delete(force=True)

@pytest.fixture(scope="module")
def test_file(storage):
    path = upload_test_file(storage, TEST_FILE_PATH, TEST_FILE_CONTENT)
    yield path
    storage.delete(path)

@pytest.fixture(scope="module")
def gcs_settings(request, storage):
    bucket_name = storage.bucket.name

    from django.test import override_settings
    with override_settings(
        GCS_PROJECT=request.config.getoption("--gcs-project-name"),
        GCS_CREDENTIALS_FILE_PATH=request.config.getoption("--gcs-credentials-file"),
        GCS_BUCKET=bucket_name
    ):
        yield True



# coding=utf-8
from __future__ import unicode_literals

import datetime
import ssl
from tempfile import TemporaryFile

import google.cloud.exceptions
import pytest
from django.core.exceptions import SuspiciousFileOperation
from django.utils import six

from django_gcloud_storage import safe_join, remove_prefix, GCloudFile

from conftest import TEST_FILE_CONTENT, TEST_FILE_PATH
from helpers import upload_test_file

def urlopen(*args, **kwargs):
    try:
        from urllib.request import urlopen
    except ImportError:
        from urllib2 import urlopen

    try:
        # Ignore SSL errors (won't work on Py3.3 but can be ignored there)
        kwargs["context"] = ssl._create_unverified_context()
    except AttributeError:  # Py3.3
        pass

    return urlopen(*args, **kwargs)


# noinspection PyClassHasNoInit,PyMethodMayBeStatic
class TestSafeJoin:
    def test_should_join_urls(self):
        assert safe_join("test", "index.html") == "test/index.html"

    def test_should_not_break_on_slash_on_base(self):
        assert safe_join("test/", "index.html") == "test/index.html"
        assert safe_join("test///", "index.html") == "test/index.html"

    def test_should_enforce_no_starting_slash_on_base(self):
        assert safe_join("/test", "index.html") == "test/index.html"
        assert safe_join("////test", "index.html") == "test/index.html"

    def test_should_resolve_dots_to_absolute_path(self):
        assert safe_join("test", "/test/../index.html") == "test/index.html"

    def test_should_resolve_multiple_slashes(self):
        assert safe_join("test", "/test//abc////index.html") == "test/test/abc/index.html"
        assert safe_join("test///", "///test//abc////index.html") == "test/test/abc/index.html"

    def test_should_not_allow_escaping_base_path(self):
        with pytest.raises(SuspiciousFileOperation):
            safe_join("test", "../index.html")
        with pytest.raises(SuspiciousFileOperation):
            safe_join("test", "/../index.html")

    def test_should_work_with_bytes(self):
        assert safe_join(b"test", "index.html") == "test/index.html"
        assert safe_join("test", b"index.html") == "test/index.html"
        assert safe_join(b"test", b"index.html") == "test/index.html"

    def test_should_work_with_unicode_characters(self):
        assert safe_join("test", "brathähnchen.html") == "test/brathähnchen.html"

    def test_should_normalize_system_dependant_slashes(self):
        assert safe_join("test", "windows\\slashes") == "test/windows/slashes"
        assert safe_join("test", "windows\\/slashes") == "test/windows/slashes"
        assert safe_join("windows\\", "slashes") == "windows/slashes"


def test_remove_prefix_function():
    assert remove_prefix("/a/b/c/", "/a/") == "b/c/"
    assert remove_prefix("/a/b/c/", "/b/") == "/a/b/c/"


# noinspection PyMethodMayBeStatic,PyTypeChecker
class TestGCloudFile:
    TEST_CONTENT = "Brathähnchen".encode("utf8")

    def test_should_be_able_to_read_and_write(self, monkeypatch):
        monkeypatch.setattr(GCloudFile, "_update_blob", lambda: None)

        f = GCloudFile(None)
        f.open("w")
        assert f.read() == (b"" if six.PY3 else "")
        f.write(self.TEST_CONTENT)
        f.seek(0)
        assert f.read() == self.TEST_CONTENT

    def test_small_temporary_files_should_not_be_rolled_over_to_disk(self, monkeypatch):
        monkeypatch.setattr(GCloudFile, "_update_blob", lambda: None)

        f = GCloudFile(None, maxsize=1000)
        f.write("a".encode("utf8") * 1000)

        assert not f._tmpfile._rolled

    def test_large_temporary_files_should_be_rolled_over_to_disk(self, monkeypatch):
        monkeypatch.setattr(GCloudFile, "_update_blob", lambda: None)

        f = GCloudFile(None, maxsize=1000)
        f.write("a".encode("utf8") * 1001)

        assert f._tmpfile._rolled

    def test_modified_files_should_be_marked_as_dirty(self, monkeypatch):
        monkeypatch.setattr(GCloudFile, "_update_blob", lambda: None)

        f = GCloudFile(None)
        f.write(self.TEST_CONTENT)

        assert f._dirty


# noinspection PyClassHasNoInit,PyMethodMayBeStatic
class TestGCloudStorageClass:
    def test_should_create_blob_at_correct_path(self, storage, test_file):
        assert test_file == TEST_FILE_PATH

    def test_should_create_a_valid_client_object(self, storage):
        with pytest.raises(google.cloud.exceptions.NotFound):
            storage.client.get_bucket("some_random_bucket_name_that_doesnt_exist")

    def test_should_create_a_valid_bucket_object(self, storage):
        assert storage.bucket.exists()

    def test_should_be_able_to_save_and_open_files(self, storage, test_file):
        f = storage.open(test_file)
        assert f.read() == TEST_FILE_CONTENT

    def test_should_return_created_time(self, storage, test_file):
        assert isinstance(storage.created_time(test_file), datetime.datetime)

    def test_should_return_modified_time(self, storage, test_file):
        assert isinstance(storage.modified_time(test_file), datetime.datetime)
        assert isinstance(storage.get_modified_time(test_file), datetime.datetime)

    def test_should_be_able_to_delete_files(self, storage):
        file_name = "test_delete_file"
        upload_test_file(storage, file_name, "")
        storage.delete(file_name)

        # Should not raise an exception by google.cloud
        assert storage.delete("missing_file") is None

    def test_exists_method(self, storage, test_file):
        assert not storage.exists("missing_file")
        assert storage.exists(test_file)

    def test_should_return_correct_file_size(self, storage, test_file):
        assert storage.size(test_file) == len(TEST_FILE_CONTENT)

    def test_should_return_publicly_downloadable_url(self, storage, test_file):
        assert urlopen(storage.url(test_file)).read() == TEST_FILE_CONTENT

    def test_should_work_with_utf8(self, storage):
        file_name = "test_utf8_file_陰陽"
        upload_test_file(storage, file_name, "")

        storage.exists(file_name)

        # Don't explode when trying to find a available name for existing files...
        upload_test_file(storage, file_name, "")

    def test_should_be_able_to_list_dirs_and_files(self, storage):
        file_name = "test_listdir_file"
        subdir_file_pattern = "/subdir/%s.%d"

        for i in range(1, 11):
            upload_test_file(storage, subdir_file_pattern % (file_name, i), "")

        upload_test_file(storage, "/subdir/a/" + file_name, "")
        upload_test_file(storage, "/subdir/b/" + file_name, "")

        # Make sure paths prefixed with a slash are normalized
        assert storage.listdir("") == storage.listdir("/")
        assert storage.listdir("subdir") == storage.listdir("/subdir")

        subdir_list_dir = storage.listdir("subdir/")
        assert len(subdir_list_dir[0]) == 2 and len(subdir_list_dir[1]) == 10
        assert subdir_list_dir[0] == ["a", "b"]
        assert subdir_list_dir[1][0] == "%s.%d" % (file_name, 1)

    def test_should_not_overwrite_files_on_save(self, storage, test_file):
        duplicate_file = upload_test_file(storage, test_file, "")
        assert duplicate_file != test_file

    def test_changed_files_should_be_reuploaded(self, storage):
        file_name = "test_changed_file_reuploaded"
        file_content = "gcloud".encode("ascii")

        upload_test_file(storage, file_name, "")
        first_modified_time = storage.modified_time(file_name)
        local_tmpfile = storage.open(file_name)

        assert local_tmpfile.read() == "".encode("ascii")
        local_tmpfile.seek(0)

        local_tmpfile.write(file_content)
        local_tmpfile.close()

        assert storage.open(file_name).read() == file_content
        assert storage.modified_time(file_name) != first_modified_time

    def test_open_should_be_able_to_create_new_file(self, storage):
        file_name = "test_open_creates_file"
        file_content = "".encode("ascii")

        with storage.open(file_name, mode="wb") as f:
            f.write(file_content)

        assert storage.exists(file_name)

    def test_use_unsigned_urls_option(self, storage, test_file):
        storage.use_unsigned_urls = True
        url = storage.url(test_file)
        storage.use_unsigned_urls = False
        for i in ["Signature=", "GoogleAccessId=", "Expires="]:
            assert i not in url



# -*- encoding: utf-8 -*-
from __future__ import unicode_literals

import datetime
import os
import re
from tempfile import SpooledTemporaryFile

import django
from django.conf import settings
from django.core.exceptions import SuspiciousFileOperation
from django.core.files.base import File
from django.core.files.storage import Storage
from django.utils.deconstruct import deconstructible
from django.utils.encoding import force_text, smart_str
from google.cloud import _helpers as gcloud_helpers
from google.cloud import storage
from google.cloud.exceptions import NotFound
from google.cloud.storage.bucket import Bucket

__version__ = '0.3.1'

DJANGO_17 = django.get_version().startswith('1.7.')

try:
    # For Python 3.0 and later
    from urllib import parse as urlparse
except ImportError:
    # Fall back to Python 2's urllib2
    import urlparse


def safe_join(base, path):
    base = force_text(base).replace("\\", "/").lstrip("/").rstrip("/") + "/"
    path = force_text(path).replace("\\", "/").lstrip("/")

    # Ugh... there must be a better way that I can't think of right now
    if base == "/":
        base = ""

    resolved_url = urlparse.urljoin(base, path)

    resolved_url = re.sub("//+", "/", resolved_url)

    if not resolved_url.startswith(base):
        raise SuspiciousFileOperation(
            'The joined path ({}) is located outside of the base path '
            'component ({})'.format(resolved_url, base))

    return resolved_url


def prepare_name(name):
    return smart_str(name, encoding='utf-8')


def remove_prefix(target, prefix):
    if target.startswith(prefix):
        return target[len(prefix):]
    return target


class GCloudFile(File):
    """
    Django file object that wraps a SpooledTemporaryFile and remembers changes on
    write to reupload the file to GCS on close()
    """

    def __init__(self, blob, maxsize=1000):
        """
        :type blob: google.cloud.storage.blob.Blob
        """
        self._dirty = False
        self._tmpfile = SpooledTemporaryFile(
            max_size=maxsize,
            prefix="django_gcloud_storage_"
        )

        self._blob = blob

        super(GCloudFile, self).__init__(self._tmpfile)

    def _update_blob(self):
        # Specify explicit size to avoid problems with not yet spooled temporary files
        # Djangos File.size property already knows how to handle cases like this

        if DJANGO_17 and self._tmpfile.name is None:  # Django bug #22307
            size = self._tmpfile.tell()
        else:
            size = self.size

        self._blob.upload_from_file(self._tmpfile, size=size, rewind=True)

    def write(self, content):
        self._dirty = True
        super(GCloudFile, self).write(content)

    def close(self):
        if self._dirty:
            self._update_blob()
            self._dirty = False

        super(GCloudFile, self).close()


# noinspection PyAbstractClass
@deconstructible
class DjangoGCloudStorage(Storage):

    def __init__(self, project=None, bucket=None, credentials_file_path=None, use_unsigned_urls=None):
        self._client = None
        self._bucket = None

        if bucket is not None:
            self.bucket_name = bucket
        else:
            self.bucket_name = settings.GCS_BUCKET

        if credentials_file_path is not None:
            self.credentials_file_path = credentials_file_path
        else:
            self.credentials_file_path = settings.GCS_CREDENTIALS_FILE_PATH

        assert os.path.exists(self.credentials_file_path), "Credentials file not found"

        if project is not None:
            self.project_name = project
        else:
            self.project_name = settings.GCS_PROJECT

        if use_unsigned_urls is not None:
            self.use_unsigned_urls = use_unsigned_urls
        else:
            self.use_unsigned_urls = getattr(settings, "GCS_USE_UNSIGNED_URLS", False)

        self.bucket_subdir = ''  # TODO should be a parameter

    @property
    def client(self):
        """
        :rtype: storage.Client
        """
        if not self._client:
            self._client = storage.Client.from_service_account_json(
                self.credentials_file_path,
                project=self.project_name
            )
        return self._client

    @property
    def bucket(self):
        """
        :rtype: Bucket
        """
        if not self._bucket:
            self._bucket = self.client.get_bucket(self.bucket_name)
        return self._bucket

    def _save(self, name, content):
        name = safe_join(self.bucket_subdir, name)
        name = prepare_name(name)

        # Required for InMemoryUploadedFile objects, as they have no fileno
        total_bytes = None if not hasattr(content, 'size') else content.size

        blob = self.bucket.blob(name)
        blob.upload_from_file(content, size=total_bytes)

        return name

    def _open(self, name, mode):
        # TODO implement mode?

        name = safe_join(self.bucket_subdir, name)
        name = prepare_name(name)

        blob = self.bucket.get_blob(name)
        if blob is None:
            # Create new
            blob = self.bucket.blob(name)
            tmpfile = GCloudFile(blob)
        else:
            tmpfile = GCloudFile(blob)
            blob.download_to_file(tmpfile)
        tmpfile.seek(0)

        return tmpfile

    def created_time(self, name):
        name = safe_join(self.bucket_subdir, name)
        name = prepare_name(name)

        blob = self.bucket.get_blob(name)

        # google.cloud doesn't provide a public method for this
        value = blob._properties.get("timeCreated", None)
        if value is not None:
            naive = datetime.datetime.strptime(value, gcloud_helpers._RFC3339_MICROS)
            return naive.replace(tzinfo=gcloud_helpers.UTC)

    def delete(self, name):
        name = safe_join(self.bucket_subdir, name)
        name = prepare_name(name)

        try:
            self.bucket.delete_blob(name)
        except NotFound:
            pass

    def exists(self, name):
        name = safe_join(self.bucket_subdir, name)
        name = prepare_name(name)

        return self.bucket.get_blob(name) is not None

    def size(self, name):
        name = safe_join(self.bucket_subdir, name)
        name = prepare_name(name)

        blob = self.bucket.get_blob(name)

        return blob.size if blob is not None else None

    def modified_time(self, name):
        name = safe_join(self.bucket_subdir, name)
        name = prepare_name(name)

        blob = self.bucket.get_blob(name)

        return blob.updated if blob is not None else None

    def get_modified_time(self, name):
        # In Django>=1.10, modified_time is deprecated, and modified_time will be removed in Django 2.0.
        return self.modified_time(name)

    def listdir(self, path):
        path = safe_join(self.bucket_subdir, path)
        path = prepare_name(path)

        iterator = self.bucket.list_blobs(
            prefix=path,
            delimiter="/"
        )

        items = [remove_prefix(blob.name, path) for blob in list(iterator)]
        # prefixes is only set after first iterating the results!
        dirs = [remove_prefix(prefix, path).rstrip("/") for prefix in list(iterator.prefixes)]

        items.sort()
        dirs.sort()

        return dirs, items

    def url(self, name):
        name = safe_join(self.bucket_subdir, name)
        name = prepare_name(name)

        if self.use_unsigned_urls:
          return "https://storage.googleapis.com/{}/{}".format(self.bucket.name, name)

        return self.bucket.get_blob(name).generate_signed_url(expiration=datetime.datetime.now() + datetime.timedelta(hours=1))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def get_version(*file_paths):
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    with open(filename) as f:
        version_file = f.read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')

version = get_version('django_gcloud_storage', '__init__.py')

with open('README.rst') as f:
    readme = f.read()
with open('HISTORY.rst') as f:
    history = f.read().replace('.. :changelog:', '')

setup(
    name='django-gcloud-storage',
    version=version,
    description="""Django storage module implementation for Google Cloud Storage""",
    long_description=readme + '\n\n' + history,
    author='Sven Grunewaldt',
    author_email='strayer@olle-orks.org',
    url='https://github.com/strayer/django-gcloud-storage',
    packages=[
        'django_gcloud_storage',
    ],
    include_package_data=True,
    install_requires=[
        "google-cloud-storage>=1.10.0",
        "django>=1.11"
    ],
    license="BSD",
    zip_safe=False,
    keywords='django-gcloud-storage gcloud google-cloud gcs',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)
